from typing import Any, Dict

RUNTIME_DEFAULTS = {
    "ENGINE": "pandas",
    "OUTDIR": "output",
}


class RunConfig:
    def __init__(self, options: Dict[str, Any], project: Dict[str, Any]) -> None:
        options = {k: v for k, v in options.items() if v is not None}

        def get_final_value(key):
            return options.get(key, project.get(key, RUNTIME_DEFAULTS[key.upper()]))

        self.engine = get_final_value("engine")
        self.outdir = get_final_value("outdir")



from sqlalchemy import MetaData, Table, inspect


def get_schemas_from_db(engine):
    inspector = inspect(engine)
    return sorted(inspector.get_schema_names())


def get_tables_from_db(engine):
    inspector = inspect(engine)
    return sorted(inspector.get_table_names())


def get_tables_from_schema(engine, schema):
    inspector = inspect(engine)
    return sorted(inspector.get_table_names(schema=schema))


def get_table_primary_keys(engine, table):
    tabmeta = Table(table, MetaData(), autoload_with=engine)
    return tabmeta.primary_key.columns.values()



from pathlib import Path
from typing import Union


def path_exists(p: Union[str, Path]) -> bool:
    return Path(p).exists()


def recursive_rmdir(directory):
    directory = Path(directory)
    for item in directory.iterdir():
        if item.is_dir():
            recursive_rmdir(item)
        else:
            item.unlink()
    directory.rmdir()


def create_or_replace_dir(d: Union[str, Path]) -> Path:
    p = Path(d)
    if p.exists():
        recursive_rmdir(p)
    p.mkdir()
    return p


def create_dir_if_not_exist(d: Union[str, Path]) -> Path:
    p = Path(d)
    p.mkdir(parents=True, exist_ok=True)
    return p


def get_output_base_dir(base: str) -> Path:
    return create_or_replace_dir(base)


def get_result_dir(dir_dict: dict, base: Union[str, Path], key: str) -> Path:
    p = Path(get_output_base_dir(base), dir_dict[key])
    return create_or_replace_dir(p)



from typing import Dict


def extract_table_name_from_config(config: Dict, dbtype: str):
    dbtype = dbtype.lower()
    if dbtype in ("snowflake", "mssql", "postgres"):
        table = f"{config['database']}.{config['schema']}.{config['table']}"
    elif dbtype == "mysql":
        table = f"{config['schema']}.{config['table']}"

    return table



from typing import Dict

from sqlalchemy import create_engine
from sqlalchemy.engine import URL

from tulona.exceptions import TulonaMissingPropertyError


def get_mssql_engine(conn_profile: Dict):
    if "connection_string" in conn_profile:
        connection_string = conn_profile["connection_string"]
        url = URL.create("mssql+pyodbc", query={"odbc_connect": connection_string})
        engine = create_engine(url)

    # validate properties
    if "connection_string" not in conn_profile:
        mandaory_properties = {"driver_version", "server", "database"}
        if len(mandaory_properties.intersection(set(conn_profile.keys()))) != len(
            mandaory_properties
        ):
            raise TulonaMissingPropertyError(
                f"One or more of {mandaory_properties} connection propertie[s] is/are missing"
            )

    return engine



import logging
import time
from dataclasses import _MISSING_TYPE, dataclass, fields
from datetime import datetime
from pathlib import Path
from typing import Dict, List

import pandas as pd

from tulona.config.runtime import RunConfig
from tulona.exceptions import TulonaMissingPrimaryKeyError, TulonaMissingPropertyError
from tulona.task.base import BaseTask
from tulona.util.dataframe import apply_column_exclusion
from tulona.util.excel import highlight_mismatch_cells
from tulona.util.filesystem import create_dir_if_not_exist
from tulona.util.profiles import extract_profile_name, get_connection_profile
from tulona.util.project import extract_table_name_from_config
from tulona.util.sql import (
    build_filter_query_expression,
    get_column_query,
    get_query_output_as_df,
    get_sample_row_query,
)

log = logging.getLogger(__name__)

DEFAULT_VALUES = {
    "sample_count": 20,
}


@dataclass
class CompareDataTask(BaseTask):
    profile: Dict
    project: Dict
    runtime: RunConfig
    datasources: List[str]
    sample_count: int = DEFAULT_VALUES["sample_count"]

    # Support for default values
    def __post_init__(self):
        for field in fields(self):
            # If there is a default and the value of the field is none we can assign a value
            if (
                not isinstance(field.default, _MISSING_TYPE)
                and getattr(self, field.name) is None
            ):
                setattr(self, field.name, field.default)

    # TODO: needs refactoring to remove duplicate some code
    def get_table_data(self, datasource, query_expr: str = None):
        connection_profile = get_connection_profile(
            self.profile, self.project, datasource
        )
        conman = self.get_connection_manager(conn_profile=connection_profile)

        ds_dict = self.project["datasources"][datasource]
        dbtype = self.profile["profiles"][extract_profile_name(self.project, datasource)][
            "type"
        ]
        table_name = extract_table_name_from_config(config=ds_dict, dbtype=dbtype)

        if query_expr:
            query = f"select * from {table_name} where {query_expr}"
        else:
            query = get_sample_row_query(
                dbtype=dbtype, table_name=table_name, sample_count=self.sample_count
            )

        df = get_query_output_as_df(connection_manager=conman, query_text=query)
        return df

    def get_outfile_fqn(self, ds_list):
        outdir = create_dir_if_not_exist(self.project["outdir"])
        out_timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        outfile = f"{'_'.join(ds_list)}_data_comparison_{out_timestamp}.xlsx"
        outfile_fqn = Path(outdir, outfile)
        return outfile_fqn

    def execute(self):
        log.info("Starting task: Compare")
        start_time = time.time()

        if len(self.datasources) != 2:
            raise ValueError("Comparison works between two entities, not more, not less.")

        datasource1, datasource2 = self.datasources
        ds_dict1 = self.project["datasources"][datasource1]
        ds_dict2 = self.project["datasources"][datasource2]

        dbtype1 = self.profile["profiles"][
            extract_profile_name(self.project, datasource1)
        ]["type"]
        dbtype2 = self.profile["profiles"][
            extract_profile_name(self.project, datasource2)
        ]["type"]
        table_name1 = extract_table_name_from_config(config=ds_dict1, dbtype=dbtype1)
        table_name2 = extract_table_name_from_config(config=ds_dict2, dbtype=dbtype2)

        # Extract rows from both data sources
        log.debug(
            f"Trying to extract {self.sample_count} common records from both data sources"
        )
        if "primary_key" in ds_dict1 and "primary_key" in ds_dict2:
            i = 0
            while i < 10:
                log.debug(f"Extraction iteration: {i + 1}")

                df1 = self.get_table_data(datasource=datasource1)
                if df1.shape[0] == 0:
                    raise ValueError(f"Table {table_name1} doesn't have any data")

                df1 = df1.rename(columns={c: c.lower() for c in df1.columns})

                if ds_dict1["primary_key"].lower() not in df1.columns.tolist():
                    raise ValueError(
                        f"Primary key {ds_dict1['primary_key'].lower()} not present in {table_name1}"
                    )

                df2 = self.get_table_data(
                    datasource=datasource2,
                    query_expr=build_filter_query_expression(
                        df1, ds_dict1["primary_key"].lower()
                    ),
                )

                df2 = df2.rename(columns={c: c.lower() for c in df2.columns})

                if ds_dict2["primary_key"].lower() not in df2.columns.tolist():
                    raise ValueError(
                        f"Primary key {ds_dict2['primary_key'].lower()} not present in {table_name2}"
                    )

                if df2.shape[0] > 0:
                    df1 = df1[
                        df1[ds_dict1["primary_key"].lower()].isin(
                            df2[ds_dict1["primary_key"].lower()].tolist()
                        )
                    ]
                    break

                else:
                    datasource2, datasource1 = self.datasources
                    dbtype1 = self.profile["profiles"][
                        extract_profile_name(self.project, datasource1)
                    ]["type"]
                    dbtype2 = self.profile["profiles"][
                        extract_profile_name(self.project, datasource2)
                    ]["type"]
                    table_name1 = extract_table_name_from_config(
                        config=ds_dict1, dbtype=dbtype1
                    )
                    table_name2 = extract_table_name_from_config(
                        config=ds_dict2, dbtype=dbtype2
                    )

                i += 1

            if df1.shape[0] == 0:
                raise ValueError(
                    f"Could not find common data between {table_name1} and {table_name2}"
                )
        else:
            raise TulonaMissingPrimaryKeyError(
                "Primary key is required for data comparison"
            )

        # Exclude columns
        log.debug("Excluding columns")
        if "exclude_columns" in ds_dict1:
            df1 = apply_column_exclusion(df1, ds_dict1, table_name1)
        if "exclude_columns" in ds_dict2:
            df2 = apply_column_exclusion(df2, ds_dict2, table_name2)

        # Compare
        common_columns = list(
            set(df1.columns)
            .intersection(set(df2.columns))
            .union({ds_dict1["primary_key"].lower()})
            .union({ds_dict2["primary_key"].lower()})
        )
        df1 = df1[common_columns].rename(
            columns={c: c + "_" + datasource1.replace("_", "") for c in df1.columns}
        )
        df2 = df2[common_columns].rename(
            columns={c: c + "_" + datasource2.replace("_", "") for c in df2.columns}
        )

        ds1_compressed = datasource1.replace("_", "")
        ds2_compressed = datasource2.replace("_", "")

        df_merge = pd.merge(
            left=df1 if i % 2 == 0 else df2,
            right=df2 if i % 2 == 0 else df1,
            left_on=ds_dict1["primary_key"].lower() + "_" + ds1_compressed,
            right_on=ds_dict2["primary_key"].lower() + "_" + ds2_compressed,
            validate="one_to_one",
        )

        df_merge = df_merge[sorted(df_merge.columns.tolist())]

        outfile_fqn = self.get_outfile_fqn([ds1_compressed, ds2_compressed])
        log.debug("Writing comparison result into: {outfile_fqn}")
        df_merge.to_excel(outfile_fqn, sheet_name="Data Comparison", index=False)

        log.debug("Highlighting mismtach cells")
        highlight_mismatch_cells(
            excel_file=outfile_fqn, sheet="Data Comparison", num_ds=len(self.datasources)
        )

        end_time = time.time()
        log.info("Finished task: Compare")
        log.info(f"Total time taken: {(end_time - start_time):.2f} seconds")


@dataclass
class CompareColumnTask(BaseTask):
    profile: Dict
    project: Dict
    runtime: RunConfig
    datasources: List[str]

    def get_column_data(self, datasource, table, column):
        connection_profile = get_connection_profile(
            self.profile, self.project, datasource
        )
        conman = self.get_connection_manager(conn_profile=connection_profile)

        query = get_column_query(table, column)
        try:
            log.debug(f"Trying unquoted column name: {column}")
            df = get_query_output_as_df(connection_manager=conman, query_text=query)
        except Exception as exp:
            log.debug(f"Failed with error: {exp}")
            log.debug(f'Trying quoted column name: "{column}"')
            query = get_column_query(table, column, quoted=True)
            df = get_query_output_as_df(connection_manager=conman, query_text=query)
        return df

    def write_result(self, df, ds1, ds2):
        outdir = create_dir_if_not_exist(self.project["outdir"])
        out_timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        outfile = f"{ds1}_{ds2}_column_comparison_{out_timestamp}.xlsx"
        outfile_fqn = Path(outdir, outfile)

        log.debug(f"Writing output into: {outfile_fqn}")
        df.to_excel(outfile_fqn, sheet_name="Column Comparison", index=False)

    def execute(self):
        log.info("Starting task: compare-column")
        start_time = time.time()

        if len(self.datasources) != 2:
            raise ValueError("Comparison works between two entities, not more, not less.")

        datasource1, datasource2 = self.datasources
        if ":" in datasource1 and ":" in datasource2:
            datasource1, column1 = datasource1.split(":")
            datasource2, column2 = datasource2.split(":")
        elif ":" in datasource1:
            datasource1, column1 = datasource1.split(":")
            column2 = column1
        elif ":" in datasource2:
            datasource2, column2 = datasource2.split(":")
            column1 = column2
        elif (
            "compare_column" in self.project["datasources"][datasource1]
            and "compare_column" in self.project["datasources"][datasource2]
        ):
            column1 = self.project["datasources"][datasource1]["compare_column"]
            column2 = self.project["datasources"][datasource2]["compare_column"]
        elif "compare_column" in self.project["datasources"][datasource1]:
            column1 = self.project["datasources"][datasource1]["compare_column"]
            column2 = column1
        elif "compare_column" in self.project["datasources"][datasource2]:
            column2 = self.project["datasources"][datasource2]["compare_column"]
            column1 = column2
        else:
            raise TulonaMissingPropertyError(
                "Column name must be specified for task: compare-column"
                " either by specifying 'compare_column' property in"
                " at least one of the datasource[project] configs"
                " (check sample tulona-project.yml file for example)"
                " or with '--datasources' command line argument"
                " using one of the following formats"
                " (column name is same for option 3 and 4):-"
                " 1. <datasource1>:<col1>,<datasource2>:<col2>"
                " 2. <datasource1>:<col>,<datasource2>:<col>"
                " 3. <datasource1>:<col>,<datasource2>"
                " 4. <datasource1>,<datasource2>:<col>"
            )

        ds_dict1 = self.project["datasources"][datasource1]
        ds_dict2 = self.project["datasources"][datasource2]

        dbtype1 = self.profile["profiles"][
            extract_profile_name(self.project, datasource1)
        ]["type"]
        dbtype2 = self.profile["profiles"][
            extract_profile_name(self.project, datasource2)
        ]["type"]
        table_name1 = extract_table_name_from_config(config=ds_dict1, dbtype=dbtype1)
        table_name2 = extract_table_name_from_config(config=ds_dict2, dbtype=dbtype2)

        log.debug(f"Extracting data from table: {table_name1}")
        df1 = self.get_column_data(datasource1, table_name1, column1)
        log.debug(f"Extracting data from table: {table_name2}")
        df2 = self.get_column_data(datasource2, table_name2, column2)

        df1 = df1.rename(columns={c: c.lower() for c in df2.columns})
        df2 = df2.rename(columns={c: c.lower() for c in df2.columns})
        column1, column2 = column1.lower(), column2.lower()

        ds1_compressed = datasource1.replace("_", "")
        ds2_compressed = datasource2.replace("_", "")

        df_merge = pd.merge(
            left=df1,
            right=df2,
            left_on=column1,
            right_on=column2,
            how="outer",
            suffixes=["_left_" + ds1_compressed, "_right_" + ds2_compressed],
            validate="one_to_one",
            indicator="presence",
        )
        df_merge = df_merge[df_merge["presence"] != "both"]

        log.debug("Writing comparison result")
        self.write_result(df_merge, ds1_compressed, ds2_compressed)

        end_time = time.time()
        log.info("Finished task: compare-column")
        log.info(f"Total time taken: {(end_time - start_time):.2f} seconds")



import logging

import click

from tulona.cli import params as p
from tulona.config.profile import Profile
from tulona.config.project import Project
from tulona.config.runtime import RunConfig
from tulona.exceptions import TulonaMissingArgumentError
from tulona.task.compare import CompareColumnTask, CompareDataTask

# from tulona.task.scan import ScanTask
from tulona.task.ping import PingTask
from tulona.task.profile import ProfileTask

log = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.DEBUG,  # TODO: Set level to INFO once verbosity is fixed
    format="[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
)

# TODO: Make use of command line arguments like exec_engine, outdir etc.
# to override project config values.


# command: tulona
@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    epilog="Execute: tulona <command> -h/--help for more help with specific commands",
)
@click.pass_context
def cli(ctx):
    """Tulona compares data sources to find out differences"""


# command: tulona ping
@cli.command("ping")
@click.pass_context
@p.exec_engine
@p.outdir
@p.verbose
@p.datasources
def ping(ctx, **kwargs):
    """Test connectivity to datasources"""

    if "datasources" not in kwargs:
        raise TulonaMissingArgumentError(
            "--datasources argument must be provided with command: ping"
        )

    if kwargs["verbose"]:
        # TODO: Fix me
        # This setting doesn't enable debug level logging
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]

    datasource_list = kwargs["datasources"].split(",")

    task = PingTask(ctx.obj["profile"], ctx.obj["project"], datasource_list)
    task.execute()


# # command: tulona scan
# @cli.command("scan")
# @click.pass_context
# @p.exec_engine
# @p.outdir
# @p.verbose
# @p.datasources
# def scan(ctx, **kwargs):
#     """Scans data sources for schemas, tables, columns etc."""

#     if "datasources" not in kwargs:
#         raise TulonaMissingArgumentError(
#             "--datasources argument must be provided with command: scan"
#         )

#     if kwargs["verbose"]:
#         # TODO: Fix me
#         # This setting doesn't enable debug level logging
#         handler = logging.StreamHandler()
#         handler.setLevel(logging.DEBUG)

#     prof = Profile()
#     proj = Project()

#     ctx.obj = ctx.obj or {}
#     ctx.obj["project"] = proj.load_project_config()
#     ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
#     ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

#     datasource_list = kwargs["datasources"].split(",")

#     task = ScanTask(ctx.obj["profile"], ctx.obj["project"], ctx.obj["runtime"], datasource_list)
#     task.execute()


# command: tulona profile
@cli.command("profile")
@click.pass_context
@p.exec_engine
@p.outdir
@p.verbose
@p.datasources
@p.compare
def profile(ctx, **kwargs):
    """Profile data sources to collect metadata [row count, column min/max/mean etc.]"""

    if "datasources" not in kwargs:
        raise TulonaMissingArgumentError(
            "--datasources argument must be provided with command: profile"
        )

    if kwargs["verbose"]:
        # TODO: Fix me
        # This setting doesn't enable debug level logging
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    datasource_list = kwargs["datasources"].split(",")

    task = ProfileTask(
        ctx.obj["profile"],
        ctx.obj["project"],
        ctx.obj["runtime"],
        datasource_list,
        compare=kwargs["compare"],
    )
    task.execute()


# command: tulona compare-data
@cli.command("compare-data")
@click.pass_context
@p.exec_engine
@p.outdir
@p.verbose
@p.datasources
@p.sample_count
def compare_data(ctx, **kwargs):
    """Compares two data entities"""

    if "datasources" not in kwargs:
        raise TulonaMissingArgumentError(
            "--datasources argument must be provided with command: compare-data"
        )

    if kwargs["verbose"]:
        # TODO: Fix me
        # This setting doesn't enable debug level logging
        logging.basicConfig(
            level=logging.DEBUG,
            format="[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
        )

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    datasource_list = kwargs["datasources"].split(",")

    task = CompareDataTask(
        profile=ctx.obj["profile"],
        project=ctx.obj["project"],
        runtime=ctx.obj["runtime"],
        datasources=datasource_list,
        sample_count=kwargs["sample_count"],
    )
    task.execute()


# command: tulona compare-column
@cli.command("compare-column")
@click.pass_context
@p.exec_engine
@p.outdir
@p.verbose
@p.datasources
def compare_column(ctx, **kwargs):
    """
    Column name must be specified for task: compare-column
    either by specifying 'compare_column' property in
    at least one of the datasource[project] configs
    (check sample tulona-project.yml file for example)
    or with '--datasources' command line argument
    using one of the following formats
    (column name is same for option 3 and 4):-
    1. <datasource1>:<col1>,<datasource2>:<col2>
    2. <datasource1>:<col>,<datasource2>:<col>
    3. <datasource1>:<col>,<datasource2>
    4. <datasource1>,<datasource2>:<col>
    """

    if "datasources" not in kwargs:
        raise TulonaMissingArgumentError(
            "--datasources argument must be provided with command: compare-column"
        )

    if kwargs["verbose"]:
        # TODO: Fix me
        # This setting doesn't enable debug level logging
        logging.basicConfig(
            level=logging.DEBUG,
            format="[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
        )

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    datasource_list = kwargs["datasources"].split(",")

    task = CompareColumnTask(
        profile=ctx.obj["profile"],
        project=ctx.obj["project"],
        runtime=ctx.obj["runtime"],
        datasources=datasource_list,
    )
    task.execute()


if __name__ == "__main__":
    cli()



from typing import Dict

import pandas as pd

from tulona.exceptions import TulonaNotImplementedError


def get_sample_row_query(dbtype: str, table_name: str, sample_count: int):
    dbtype = dbtype.lower()

    # TODO: validate sampling mechanism for maximum possible randomness
    if dbtype in ["snowflake", "mssql"]:
        query = f"select * from {table_name} tablesample ({sample_count} rows)"
    elif dbtype == "postgres":
        # TODO: system_rows method not implemented, tablesample works for percentage selection
        # query = f"select * from {table_name} tablesample system_rows({sample_count})"
        query = f"select * from {table_name} limit {sample_count}"
    elif dbtype == "mysql":
        query = f"select * from {table_name} limit {sample_count}"
    else:
        raise TulonaNotImplementedError(
            f"Extracting sample rows from source type {dbtype} is not implemented."
        )

    return query


def get_column_query(table_name: str, column: str, quoted=False):
    if quoted:
        query = f"""select "{column}" from {table_name}"""
    else:
        query = f"""select {column} from {table_name}"""

    return query


def get_query_output_as_df(connection_manager, query_text: str):
    with connection_manager.engine.connect() as conn:
        df = pd.read_sql_query(query_text, conn)
    return df


def build_filter_query_expression(df: pd.DataFrame, primary_key: str):
    primary_keys = df[primary_key].tolist()

    if "int" in str(df[primary_key].dtype):
        primary_keys = [str(k) for k in primary_keys]
        query_expr = f"""{primary_key} in ({", ".join(primary_keys)})"""
    else:
        query_expr = f"""{primary_key} in ('{"', '".join(primary_keys)}')"""

    return query_expr


def get_metadata_query(database, schema, table):
    if database:
        query = f"""
        select * from information_schema.columns
        where upper(table_catalog) = '{database.upper()}'
        and upper(table_schema) = '{schema.upper()}'
        and upper(table_name) = '{table.upper()}'
        """
    else:
        query = f"""
        select * from information_schema.columns
        where upper(table_schema) = '{schema.upper()}'
        and upper(table_name) = '{table.upper()}'
        """
    return query


def get_metric_query(
    database, schema, table, columns_dtype: Dict, metrics: list, quoted=False
):
    numeric_types = [
        "smallint",
        "integer",
        "bigint",
        "decimal",
        "numeric",
        "real",
        "double precision",
        "smallserial",
        "serial",
        "bigserial",
        "tinyint",
        "mediumint",
        "int",
        "float",
        "float4",
        "float8",
        "double",
        "number",
        "byteint",
        "bit",
        "smallmoney",
        "money",
    ]
    timestamp_types = [
        "timestamp",
        "date",
        "time",
        "year",
        "datetime",
        "interval",
        "datetimeoffset",
        "smalldatetime",
        "datetime2",
        "timestamp_tz",
        "timestamp_ltz",
        "timestamp_ntz",
        "timestamp with time zone",  # TODO probably incorrect representation
        "timestamp without time zone",
    ]

    numeric_funcs = [
        "min",
        "max",
        "average",
        "avg",
    ]
    timestamp_funcs = [
        "min",
        "max",
    ]
    generic_funcs = [
        "count",
        "distinct_count",
    ]

    function_map = {
        "min": "min({}) as {}_min",
        "max": "max({}) as {}_max",
        "avg": "avg({}) as {}_avg",
        "average": "avg({}) as {}_average",
        "count": "count({}) as {}_count",
        "distinct_count": "count(distinct({})) as {}_distinct_count",
    }

    call_funcs = []
    for col, dtype in columns_dtype.items():
        dtype = dtype.lower()
        qp = []
        for m in metrics:
            if (
                (m in numeric_funcs and dtype in numeric_types)
                or (m in timestamp_funcs and dtype in timestamp_types)
                or (m in generic_funcs)
            ):
                qp.append(
                    function_map[m.lower()].format(f'"{col}"' if quoted else col, col)
                )
            else:
                qp.append(f"'NA' as {col}_{m.lower()}")
        call_funcs.extend(qp)

    table_fqn = f"{database if database else ''}{'.' if database else ''}{schema}.{table}"

    query = f"""
    select
        {", ".join(call_funcs)}
    from {table_fqn}
    """

    return query



import logging
from pathlib import Path

from tulona.exceptions import TulonaInvalidProjectConfigError, TulonaProjectException
from tulona.util.filesystem import path_exists
from tulona.util.yaml_parser import read_yaml

log = logging.getLogger(__name__)

PROJECT_FILE_NAME = "tulona-project.yml"


class Project:
    @property
    def get_project_root(self):
        return Path().absolute()

    @property
    def project_conf_path(self) -> str:
        return Path(self.get_project_root, PROJECT_FILE_NAME)

    def validate_project_config(self, project_dict_raw: dict) -> bool:
        # TODO: implement validations for project config
        valid = True

        if not valid:
            raise TulonaInvalidProjectConfigError("Project config is not valid")

    def load_project_config(self) -> None:
        project_file_uri = self.project_conf_path
        log.debug(f"Attempting to load project config from {project_file_uri}")

        if not path_exists(project_file_uri):
            raise TulonaProjectException(
                f"Project file {project_file_uri} does not exist."
            )

        project_dict_raw = read_yaml(project_file_uri)

        if not isinstance(project_dict_raw, dict):
            raise TulonaProjectException(
                f"{project_file_uri} could not be parsed to a python dictionary."
            )

        log.debug(f"Project config is successfully loaded from {project_file_uri}")

        self.validate_project_config(project_dict_raw)

        return project_dict_raw



from typing import Dict

import pandas as pd

from tulona.exceptions import TulonaFundamentalError


def apply_column_exclusion(df: pd.DataFrame, config: Dict, table: str):
    if config["primary_key"] in config["exclude_columns"]:
        raise TulonaFundamentalError(
            "Cannot exclude primary key/join key from data comparison"
        )

    missing_cols = []
    for col in config["exclude_columns"]:
        if col not in df.columns.tolist():
            missing_cols.append(col)

    if len(missing_cols) > 0:
        raise ValueError(
            f"Columns [{missing_cols}] to be excluded are not present in {table}"
        )

    df = df.drop(columns=config["exclude_columns"])
    return df



import logging
from pathlib import Path

from tulona.exceptions import TulonaInvalidProfileConfigError, TulonaProfileException
from tulona.util.filesystem import path_exists
from tulona.util.yaml_parser import read_yaml

log = logging.getLogger(__name__)

PROFILE_FOLDER_NAME = ".tulona"
PROFILE_FILE_NAME = "profiles.yml"


class Profile:
    @property
    def get_profile_root(self):
        return Path(Path.home(), ".tulona")

    @property
    def profile_conf_path(self) -> str:
        return Path(self.get_profile_root, PROFILE_FILE_NAME)

    def validate_profile_config(self, profile_dict_raw: dict) -> bool:
        # TODO: implement validations for profile config
        valid = True

        if not valid:
            raise TulonaInvalidProfileConfigError(
                "Invalid profile config. Please check 'profiles.yml'"
            )

    def load_profile_config(self) -> None:
        profile_file_uri = self.profile_conf_path
        log.debug(f"Attempting to load profile config from {profile_file_uri}")

        if not path_exists(profile_file_uri):
            raise TulonaProfileException(
                f"Profile file {profile_file_uri} does not exist."
            )

        profile_dict_raw = read_yaml(profile_file_uri)

        if not isinstance(profile_dict_raw, dict):
            raise TulonaProfileException(
                f"{profile_file_uri} could not be parsed to a python dictionary."
            )

        log.debug(f"Profile config is successfully loaded from {profile_file_uri}")

        self.validate_profile_config(profile_dict_raw)

        return profile_dict_raw



import logging
from typing import Dict

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from snowflake.sqlalchemy import URL
from sqlalchemy import create_engine

logging.getLogger("snowflake").setLevel(logging.ERROR)


def get_snowflake_engine(conn_profile: Dict):

    if "password" in conn_profile:
        engine = create_engine(
            URL(
                account=conn_profile["account"],
                warehouse=conn_profile["warehouse"],
                role=conn_profile["role"] if "role" in conn_profile else None,
                database=conn_profile["database"],
                schema=conn_profile["schema"],
                user=conn_profile["user"],
                password=conn_profile["password"],
            ),
            echo=False,
        )  # TODO: remove echo_pool="debug" param

    if "private_key" in conn_profile:

        # validate private_key
        if not conn_profile["private_key"].endswith(".p8"):
            raise ValueError(f"{conn_profile['private_key']} is not a valid private key")

        password = (
            conn_profile["private_key_passphrase"].encode()
            if "private_key_passphrase" in conn_profile
            else None
        )

        with open(conn_profile["private_key"], "rb") as key:
            p_key = serialization.load_pem_private_key(
                key.read(), password=password, backend=default_backend()
            )

        pkb = p_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        engine = create_engine(
            URL(
                account=conn_profile["account"],
                warehouse=conn_profile["warehouse"],
                database=conn_profile["database"],
                schema=conn_profile["schema"],
                user=conn_profile["user"],
            ),
            connect_args={
                "private_key": pkb,
            },
            echo=False,
        )  # TODO: remove echo_pool="debug" param

    return engine



from pathlib import Path
from typing import Dict


def profile_path() -> str:
    return Path(Path.home(), ".tulona", "profiles.yml")


def profile_exists():
    return profile_path().exists()


def extract_profile_name(project: Dict, datasource: str):
    ds_profile_name = project["datasources"][datasource]["connection_profile"]
    return ds_profile_name


def get_connection_profile(profile: Dict, project: Dict, datasource: str):
    ds_profile_name = extract_profile_name(project, datasource)
    connection_profile = profile["profiles"][ds_profile_name]
    return connection_profile



import logging
from dataclasses import dataclass

from tulona.adapter.base.connection import BaseConnectionManager
from tulona.adapter.mssql import get_mssql_engine
from tulona.adapter.mysql import get_mysql_engine
from tulona.adapter.postgres import get_postgres_engine
from tulona.adapter.snowflake import get_snowflake_engine
from tulona.exceptions import TulonaNotImplementedError

log = logging.getLogger(__name__)


@dataclass
class ConnectionManager(BaseConnectionManager):
    def get_engine(self):
        dbtype = self.conn_profile["type"].lower()
        if dbtype == "snowflake":
            self.engine = get_snowflake_engine(self.conn_profile)
        elif dbtype == "mssql":
            self.engine = get_mssql_engine(self.conn_profile)
        elif dbtype == "postgres":
            self.engine = get_postgres_engine(self.conn_profile)
        elif dbtype == "mysql":
            self.engine = get_mysql_engine(self.conn_profile)
        else:
            raise TulonaNotImplementedError(
                f"Tulona connection manager is not set up for {dbtype}"
            )

    def open(self):
        self.get_engine()
        self.conn = self.engine.connect()

    def close(self):
        self.conn.close()



from typing import Dict

from sqlalchemy import create_engine
from sqlalchemy.engine import URL

from tulona.exceptions import TulonaMissingPropertyError


def get_postgres_engine(conn_profile: Dict):
    # TODO: Implement
    # if 'connection_string' in conn_profile:
    #     connection_string = conn_profile['connection_string']
    #     url = URL.create(drivername="postgresql+psycopg2", query={"odbc_connect": connection_string})
    #     engine = create_engine(url)

    # validate properties
    if "connection_string" not in conn_profile:
        mandaory_properties = {"host", "database", "username"}
        if len(mandaory_properties.intersection(set(conn_profile.keys()))) != len(
            mandaory_properties
        ):
            raise TulonaMissingPropertyError(
                f"One or more of {mandaory_properties} connection propertie[s] is/are missing"
            )

    if "password" in conn_profile:
        url = URL.create(
            drivername="postgresql+psycopg2",
            username=conn_profile["username"],
            password=conn_profile["password"],  # plain (unescaped) text
            database=conn_profile["database"],
            host=conn_profile["host"],
            port=conn_profile["port"],
        )
        engine = create_engine(url, echo=False)

    return engine



import click

exec_engine = click.option(
    "--engine", help="Execution engine. Can be one of Pandas right now", type=click.STRING
)

outdir = click.option(
    "--outdir",
    help="Directory to write the result of the comparison and other related metadata",
    type=click.STRING,
)

verbose = click.option("--verbose", "-v", is_flag=True, help="Show debug level logs")

datasources = click.option(
    "--datasources",
    help="Comma separated list of one or more datasource names defined in tulona-conf.yml file",
)

sample_count = click.option(
    "--sample-count", help="Number of maximum records to be compared"
)

compare = click.option(
    "--compare",
    is_flag=True,
    help="Can be used with profile task to compare profiles of different data sources",
)



class TulonaNotImplementedError(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaProjectException(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaInvalidProjectConfigError(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaUnSupportedExecEngine(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaProfileException(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaInvalidProfileConfigError(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaMissingPropertyError(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaMissingArgumentError(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaMissingPrimaryKeyError(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)


class TulonaFundamentalError(Exception):
    def __init__(self, message: str):
        self.message = message
        self.formatted_message = f"ERROR: {self.message}"
        super().__init__(self.formatted_message)



from typing import Union

from openpyxl import load_workbook, styles
from openpyxl.styles import Border, Side
from openpyxl.worksheet.worksheet import Worksheet


def get_column_index(sheet: Worksheet, column: str):
    for row in sheet.iter_rows(min_row=1, max_row=1, values_only=True):
        for idx, cell_value in enumerate(row, 1):
            if cell_value == column:
                return idx
    raise ValueError(f"Column {column} could not be found in the Excel sheet.")


def highlight_mismatch_pair(excel_file, sheet):
    wb = load_workbook(excel_file)
    ws = wb[sheet]

    yellow_fill = styles.PatternFill(
        start_color="FFFFFF00", end_color="FFFFFF00", fill_type="solid"
    )

    left_border = Border(
        left=Side(border_style="thin"),
        right=Side(border_style="dotted"),
        top=Side(border_style="thin"),
        bottom=Side(border_style="thin"),
    )
    right_border = Border(
        left=Side(border_style="dotted"),
        right=Side(border_style="thin"),
        top=Side(border_style="thin"),
        bottom=Side(border_style="thin"),
    )

    for row in ws.iter_rows(
        min_row=2, min_col=0, max_row=ws.max_row, max_col=ws.max_column
    ):
        for col_idx in range(0, ws.max_column, 2):
            if row[col_idx].value != row[col_idx + 1].value:
                row[col_idx].fill = yellow_fill
                row[col_idx + 1].fill = yellow_fill

                row[col_idx].border = left_border
                row[col_idx + 1].border = right_border

    wb.save(excel_file)


def highlight_mismatch_cells(
    excel_file, sheet: str, num_ds: int, skip_columns: Union[str, list[str]] = None
):
    wb = load_workbook(excel_file)
    ws = wb[sheet]

    yellow_fill = styles.PatternFill(
        start_color="FFFFFF00", end_color="FFFFFF00", fill_type="solid"
    )

    left_border = Border(
        left=Side(border_style="thin"),
        right=Side(border_style="dotted"),
        top=Side(border_style="thin"),
        bottom=Side(border_style="thin"),
    )

    middle_border = Border(
        left=Side(border_style="dotted"),
        right=Side(border_style="dotted"),
        top=Side(border_style="thin"),
        bottom=Side(border_style="thin"),
    )

    right_border = Border(
        left=Side(border_style="dotted"),
        right=Side(border_style="thin"),
        top=Side(border_style="thin"),
        bottom=Side(border_style="thin"),
    )

    if skip_columns:
        skip_columns = skip_columns if isinstance(skip_columns, list) else [skip_columns]
        skip_idxs = [get_column_index(ws, c) - 1 for c in skip_columns]
        compareable_col_idxs = list(set(range(ws.max_column)) - set(skip_idxs))
    else:
        compareable_col_idxs = list(range(ws.max_column))

    for row in ws.iter_rows(
        min_row=2, min_col=0, max_row=ws.max_row, max_col=ws.max_column
    ):
        for col_idx in compareable_col_idxs[::num_ds]:
            equal = []
            for i in range(num_ds - 1):
                equal.append(row[col_idx + i].value == row[col_idx + i + 1].value)

            if not all(equal):
                for i in range(num_ds):
                    row[col_idx + i].fill = yellow_fill

                    if i == 0:
                        row[col_idx].border = left_border
                    elif i == num_ds - 1:
                        row[col_idx + i].border = right_border
                    else:
                        row[col_idx + i].border = middle_border

    wb.save(excel_file)



import logging
import time
from dataclasses import _MISSING_TYPE, dataclass, fields
from datetime import datetime
from pathlib import Path
from typing import Dict, List

import pandas as pd

from tulona.config.runtime import RunConfig
from tulona.task.base import BaseTask
from tulona.util.excel import highlight_mismatch_cells
from tulona.util.filesystem import create_dir_if_not_exist
from tulona.util.profiles import extract_profile_name, get_connection_profile
from tulona.util.sql import get_metadata_query, get_metric_query, get_query_output_as_df

log = logging.getLogger(__name__)


@dataclass
class ProfileTask(BaseTask):
    profile: Dict
    project: Dict
    runtime: RunConfig
    datasources: List[str]
    compare: bool = False

    # Support for default values
    def __post_init__(self):
        for field in fields(self):
            # If there is a default and the value of the field is none we can assign a value
            if (
                not isinstance(field.default, _MISSING_TYPE)
                and getattr(self, field.name) is None
            ):
                setattr(self, field.name, field.default)

    def get_outfile_fqn(self, ds_list):
        outdir = create_dir_if_not_exist(self.project["outdir"])
        out_timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        outfile = f"{'_'.join(ds_list)}_profiles_{out_timestamp}.xlsx"
        outfile_fqn = Path(outdir, outfile)
        return outfile_fqn

    def execute(self):

        log.info("Starting task: profiling")
        start_time = time.time()

        df_collection = []
        ds_name_compressed_list = []
        for ds_name in self.datasources:
            # Extract data source name from datasource:column combination
            ds_name = ds_name.split(":")[0]
            ds_name_compressed = ds_name.replace("_", "")
            ds_name_compressed_list.append(ds_name_compressed)
            log.debug(f"Extracting metadata for {ds_name}")

            ds_config = self.project["datasources"][ds_name]
            dbtype = self.profile["profiles"][
                extract_profile_name(self.project, ds_name)
            ]["type"]

            # MySQL doesn't have logical database
            if "database" in ds_config and dbtype.lower() != "mysql":
                database = ds_config["database"]
            else:
                database = None
            schema = ds_config["schema"]
            table = ds_config["table"]

            connection_profile = get_connection_profile(
                self.profile, self.project, ds_name
            )
            conman = self.get_connection_manager(conn_profile=connection_profile)

            # Extract metadata
            log.debug("Extracting metadata")
            meta_query = get_metadata_query(database, schema, table)
            log.debug(f"Executing query: {meta_query}")
            df_meta = get_query_output_as_df(
                connection_manager=conman, query_text=meta_query
            )
            df_meta = df_meta.rename(columns={c: c.lower() for c in df_meta.columns})

            # Extract metrics like min, max, avg, count, distinct count etc.
            log.debug("Extracting metrics")
            metrics = [
                "min",
                "max",
                "avg",
                "count",
                "distinct_count",
            ]
            # metrics = ["count", "distinct_count"]
            metrics = list(map(lambda s: s.lower(), metrics))
            type_dict = df_meta[["column_name", "data_type"]].to_dict(orient="list")
            columns_dtype = {
                k: v for k, v in zip(type_dict["column_name"], type_dict["data_type"])
            }

            try:
                log.debug("Trying query with unquoted column names")
                metric_query = get_metric_query(
                    database, schema, table, columns_dtype, metrics
                )
                log.debug(f"Executing query: {metric_query}")
                df_metric = get_query_output_as_df(
                    connection_manager=conman, query_text=metric_query
                )
            except Exception as exp:
                log.warn(f"Previous query failed with error: {exp}")
                log.debug("Trying query with quoted column names")
                metric_query = get_metric_query(
                    database,
                    schema,
                    table,
                    columns_dtype,
                    metrics,
                    quoted=True,
                )
                log.debug(f"Executing query: {metric_query}")
                df_metric = get_query_output_as_df(
                    connection_manager=conman, query_text=metric_query
                )

            log.debug("Converting metric data into presentable format")
            metric_dict = {m: [] for m in ["column_name"] + metrics}
            for col in df_meta["column_name"]:
                metric_dict["column_name"].append(col)
                for m in metrics:
                    try:
                        metric_value = df_metric.iloc[0][f"{col}_{m}"]
                    except Exception:
                        metric_value = df_metric.iloc[0][f"{col.lower()}_{m}"]
                    metric_dict[m].append(metric_value)

            df_metric = pd.DataFrame(metric_dict)

            # Combine meta and metric data
            df = pd.merge(left=df_meta, right=df_metric, how="inner", on="column_name")

            df_collection.append(df)

        outfile_fqn = self.get_outfile_fqn(ds_name_compressed_list)

        if self.compare:
            log.debug("Preparing metadata comparison")
            common_columns = {c.lower() for c in df_collection[0].columns.tolist()}
            df_collection_final = []
            for ds_name, df in zip(ds_name_compressed_list, df_collection):
                colset = {c.lower() for c in df.columns.tolist()}
                common_columns = common_columns.intersection(colset)

            for ds_name, df in zip(ds_name_compressed_list, df_collection):
                df = df[list(common_columns)]
                df = df.rename(
                    columns={
                        c: f"{c}_{ds_name}" if c.lower() != "column_name" else c.lower()
                        for c in df.columns
                    }
                )
                df["column_name"] = df["column_name"].str.lower()
                df_collection_final.append(df)

            df_merge = df_collection_final.pop()
            for df in df_collection_final:
                df_merge = pd.merge(
                    left=df_merge, right=df, on="column_name", how="inner"
                )
            df_merge = df_merge[sorted(df_merge.columns.tolist())]
            log.debug(f"Calculated comparison for {df_merge.shape[0]} columns")

            log.debug(f"Writing results into file: {outfile_fqn}")
            primary_key_col = df_merge.pop("column_name")
            df_merge.insert(loc=0, column="column_name", value=primary_key_col)
            df_merge.to_excel(outfile_fqn, sheet_name="Metadata Comparison", index=False)

            log.debug("Highlighting mismtach cells")
            highlight_mismatch_cells(
                excel_file=outfile_fqn,
                sheet="Metadata Comparison",
                num_ds=len(ds_name_compressed_list),
                skip_columns="column_name",
            )
        else:
            log.debug(f"Writing results into file: {outfile_fqn}")
            with pd.ExcelWriter(outfile_fqn) as writer:
                for ds_name, df in zip(ds_name_compressed_list, df_collection):
                    primary_key_col = df.pop("column_name")
                    df.insert(loc=0, column="column_name", value=primary_key_col)
                    df.to_excel(writer, sheet_name=f"{ds_name} Metadata", index=False)

        end_time = time.time()
        log.info("Finished task: profiling")
        log.info(f"Total time taken: {(end_time - start_time):.2f} seconds")



import logging
import time
from dataclasses import dataclass
from typing import Dict, List

from tulona.config.runtime import RunConfig
from tulona.task.base import BaseTask

# from tulona.util.profiles import get_connection_profile

log = logging.getLogger(__name__)


@dataclass
class ScanTask(BaseTask):
    profile: Dict
    project: Dict
    runtime: RunConfig
    datasources: List[str]

    def execute(self):

        log.info("Starting task: Scan")
        start_time = time.time()

        # # TODO: Change the implementation
        # for ds in self.datasources:
        #     log.debug(f"Testing connection to data source: {ds}")

        #     connection_profile = get_connection_profile(self.profile, self.project, ds)
        #     try:
        #         conman = self.get_connection_manager(conn_profile=connection_profile)
        #         with conman.engine.open() as connection:
        #             results = connection.execute(
        #                 "select * from information_schema"
        #             ).fetchone()
        #     except Exception as exp:
        #         log.error(f"Connection to data source {ds} failed because of: {exp}")

        end_time = time.time()
        log.info("Finished task: Scan")
        log.info(f"Total time taken: {(end_time - start_time):.2f} seconds")



import logging
import time
from dataclasses import dataclass
from typing import Dict, List

from tulona.task.base import BaseTask
from tulona.util.profiles import get_connection_profile

log = logging.getLogger(__name__)


@dataclass
class PingTask(BaseTask):
    profile: Dict
    project: Dict
    datasources: List[str]

    def execute(self):

        log.info("Starting task: Ping")
        start_time = time.time()

        for ds in self.datasources:
            log.debug(f"Testing connection to data source: {ds}")

            connection_profile = get_connection_profile(self.profile, self.project, ds)
            try:
                conman = self.get_connection_manager(conn_profile=connection_profile)
                with conman.engine.connect() as connection:
                    results = connection.execute(
                        "select * from information_schema.tables"
                    ).fetchone()
                    _ = results[0]
                    log.info("Connection successful")
            except Exception as exp:
                log.error(f"Connection to data source {ds} failed because of: {exp}")

        end_time = time.time()
        log.info("Finished task: Ping")
        log.info(f"Total time taken: {(end_time - start_time):.2f} seconds")



from abc import ABCMeta, abstractmethod
from typing import Dict

from tulona.adapter.connection import ConnectionManager
from tulona.exceptions import TulonaNotImplementedError


class BaseTask(metaclass=ABCMeta):

    def get_connection_manager(self, conn_profile: Dict) -> ConnectionManager:
        conman = ConnectionManager(conn_profile)
        conman.get_engine()
        return conman

    @abstractmethod
    def execute(self):
        raise TulonaNotImplementedError(
            "This method needs to be implemented in child class"
        )
